@@ -2,7 +2,7 @@
 //
 // LTKMisc.h
 //
-// Copyright 1996 Derek B. Noonburg
+// Copyright 1996-2002 Glyph & Cog, LLC
 //
 //========================================================================
 
@@ -13,7 +13,7 @@
 #pragma interface
 #endif
 
-#include <GString.h>
+#include "GString.h"
 
 extern void ltkError(char *msg, ...);
 
