@@ -2,7 +2,7 @@
 //
 // PSOutputDev.h
 //
-// Copyright 1996 Derek B. Noonburg
+// Copyright 1996-2002 Glyph & Cog, LLC
 //
 //========================================================================
 
@@ -23,6 +23,7 @@ class GfxPath;
 class GfxFont;
 class GfxColorSpace;
 class GfxSeparationColorSpace;
+struct PSFont16Enc;
 class PSOutCustomColor;
 
 //------------------------------------------------------------------------
@@ -63,6 +64,10 @@ public:
   // Does this device use drawChar() or drawString()?
   virtual GBool useDrawChar() { return gFalse; }
 
+  // Does this device use beginType3Char/endType3Char?  Otherwise,
+  // text in Type 3 fonts will be drawn with drawChar/drawString.
+  virtual GBool interpretType3Chars() { return gFalse; }
+
   //----- initialization and control
 
   // Start a page.
@@ -124,15 +129,27 @@ public:
   virtual void opiEnd(GfxState *state, Dict *opiDict);
 #endif
 
+  //----- Type 3 font operators
+  virtual void type3D0(GfxState *state, double wx, double wy);
+  virtual void type3D1(GfxState *state, double wx, double wy,
+		       double llx, double lly, double urx, double ury);
+
+  //----- PostScript XObjects
+  virtual void psXObject(Stream *psStream, Stream *level1Stream);
+
 private:
 
   void setupResources(Dict *resDict);
   void setupFonts(Dict *resDict);
-  void setupFont(GfxFont *font);
+  void setupFont(GfxFont *font, Dict *parentResDict);
   void setupEmbeddedType1Font(Ref *id, char *psName);
-  void setupEmbeddedType1Font(GString *fileName, char *psName);
+  void setupExternalType1Font(GString *fileName, char *psName);
   void setupEmbeddedType1CFont(GfxFont *font, Ref *id, char *psName);
   void setupEmbeddedTrueTypeFont(GfxFont *font, Ref *id, char *psName);
+  void setupExternalTrueTypeFont(GfxFont *font, char *psName);
+  void setupEmbeddedCIDType0Font(GfxFont *font, Ref *id, char *psName);
+  void setupEmbeddedCIDTrueTypeFont(GfxFont *font, Ref *id, char *psName);
+  void setupType3Font(GfxFont *font, char *psName, Dict *parentResDict);
   void setupImages(Dict *resDict);
   void setupImage(Ref id, Stream *str);
   void addProcessColor(double c, double m, double y, double k);
@@ -157,6 +174,8 @@ private:
 #endif
   void writePS(const char *fmt, ...);
   void writePSString(GString *s);
+  void writePSChar(char c);
+  GString *filterPSName(GString *name);
 
   PSLevel level;		// PostScript level (1, 2, separation)
   PSOutMode mode;		// PostScript mode (PS, EPS, form)
@@ -178,6 +197,9 @@ private:
   GString **fontFileNames;	// list of names of all embedded external fonts
   int fontFileNameLen;		// number of entries in fontFileNames array
   int fontFileNameSize;		// size of fontFileNames array
+  PSFont16Enc *font16Enc;	// encodings for substitute 16-bit fonts
+  int font16EncLen;		// number of entries in font16Enc array
+  int font16EncSize;		// size of font16Enc array
 
   double tx, ty;		// global translation
   double xScale, yScale;	// global scaling
@@ -189,13 +211,17 @@ private:
   PSOutCustomColor		// used custom colors
     *customColors;
 
+  GBool inType3Char;		// inside a Type 3 CharProc
+  GString *t3String;		// Type 3 content string
+  double t3WX, t3WY,		// Type 3 character parameters
+         t3LLX, t3LLY, t3URX, t3URY;
+  GBool t3Cacheable;		// cleared if char is not cacheable
+
 #if OPI_SUPPORT
   int opi13Nest;		// nesting level of OPI 1.3 objects
   int opi20Nest;		// nesting level of OPI 2.0 objects
 #endif
 
-  GBool type3Warning;		// only show the Type 3 font warning once
-
   GBool ok;			// set up ok?
 };
 
